import yaml
import json

from os import path as fpath
from abconfig.common import Cached


class BaseFormat:
    @staticmethod
    def reader(file):
        raise NotImplementedError

    @classmethod
    def read(cls, path):
        with open(path, 'r') as file:
            read = cls.reader(file)
            if not isinstance(read, dict):
                raise IOError('Not valid file format')
            return read


class Yaml(BaseFormat):
    @staticmethod
    def reader(file):
        return yaml.load(file, Loader=yaml.FullLoader)


class Json(BaseFormat):
    @staticmethod
    def reader(file):
        return json.load(file)


class File(Cached):
    formats = (Yaml, Json)

    def __init__(self, path):
        self.path = self.isexists(path)

    def isexists(self, path):
        if fpath.exists(path) is False:
            raise FileNotFoundError(f'{path} not found')
        return path

    def __call__(self, data=None):
        errors = list()
        for reader in self.formats:
            try:
                return reader.read(self.path)
            except Exception as e:
                errors.append(f'{reader.__name__} reader: {e}')
                continue
        error_message = '\n'.join(errors)
        raise IOError(f'{self.path} read error:\n{error_message}')



from copy import deepcopy


class Dict(dict):
    def __str__(self):
        return type(self).__name__

    def __init__(self, data=None):
        self._data = data if data else dict()

    def get(self, item=None, default=None):
        if not item:
            return self._data
        if not item in self._data:
            return default
        return self._data[item]

    def items(self):
        return self._data.items()

    def __getitem__(self, item):
        return self.get(item)

    def __add__(self, *args):
        return self._add(*args)

    def __iadd__(self, *args):
        return self._add(*args)

    def _add(self, data: dict):
        result = deepcopy(self._data)
        for k,v in self._data.items():
            v2 = data.get(k, None)
            if v2: result.update(
                {k: self._merge(k,v,v2)})
            elif not v2 and isinstance(v, type):
                result.update({k: None})
        return Dict(result)

    def _merge(self, k, source, item):
        istype = False
        eq = lambda x, y: isinstance(x, y)
        if not source or isinstance(source, type):
            eq = lambda x, y: x is y
            istype = True

        converter = (source if isinstance(source, type) else
                     (type(source) if source else lambda x: x))

        eqt = lambda x: eq(source, x)
        error = lambda: self._type_error(k,
            type(converter).__name__ if not istype else converter)

        if eqt(tuple) or eqt(list):
            if eq(item, converter): error()
            return converter(item)
        elif eqt(dict):
            if not isinstance(item, dict): error()
            return (Dict(source) + Dict(item)).get()
        return converter(item)

    def _type_error(self, k, need_type):
        raise TypeError(f'{k} value must be {need_type.__name__}')

    def __call__(self, *args):
        result = deepcopy(self)
        for func in args: result += func(result.get())
        return result


class Cache(type):
    instances = {}

    def __call__(cls, *args, **kwargs):
        if not args[0] in cls.instances:
            cls.instances[args[0]] = super(Cache, cls).__call__(*args, **kwargs)
        return cls.instances[args[0]]


class Cached(metaclass=Cache):
    """ Created once and stored in the Cache.instances dictionary,
        first positional arg is unique object id to be created
    """


from os import environ


class Env:
    def __call__(self, data, prefix=None):
        result = dict()
        update = lambda k, v: result.update({k:v}) if v else None
        for k,v in data.items():
            if isinstance(v, dict):
                update(k, self(v, self._get_prefix(prefix,k)))
            else:
                update(k,self._get_env(prefix,k))
        return result

    def _get_env(self, prefix, name):
        env_name = self._get_prefix(prefix, name)
        return environ.get(env_name.upper(), None)

    def _get_prefix(self, *args):
        clean = lambda x: True if x else False
        return '_'.join(filter(clean, args))



__version__ = '0.5.0'


from abconfig.common import Dict
from abconfig.file import File
from abconfig.env import Env


class ABConfig(Dict):
    plugins = None
    env = False
    file = None

    def __init__(self, data=None):
        if str(type(self).__name__) == 'ABConfig':
            raise NotImplementedError
        init = Dict(self._attrs() if not data else self._attrs(data))
        super().__init__(init(*self._loaders).get())

    @property
    def _loaders(self):
        result = []
        if self.plugins: result.extend(self.plugins)
        if self.file: result.append(File(self.file))
        if self.env is True: result.append(Env())
        return result

    def _attrs(self, **kwargs):
        result = {str(x): getattr(self, x) for x in type(self).__dict__.keys()
                                        if (x[:2] != '__' and
                                            x[:1] != '_' and
                                            x not in ('plugins','file', 'env'))}
        if kwargs: result.update(kwargs)
        return result
